import os
import sys
import Tigger
import numpy
import tempfile
import json
import codecs
import shlex
import shutil
import glob
import subprocess

from astLib.astWCS import WCS
from Tigger.Models import SkyModel, ModelClasses


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

with codecs.open(CONFIG, "r", "utf8") as stdr:
    cab = json.load(stdr)

junk = cab["junk"]
args = []
msname = None

sofia_file = 'sofia_parameters.par'
wstd = open(sofia_file, 'w')

wstd.write('writeCat.outputDir={:s}\n'.format(OUTPUT))
port2tigger = False
image = None
writecat = False
parameterise = False

for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue
    if name == "port2tigger":
        port2tigger = value
        continue
    if name == "steps.doWriteCat":
        writecat = value
    if name == "steps.doParameterise":
        parameterise = value
    if name == "import.inFile":
        image = value

    wstd.write('{0}={1}\n'.format(name, value))

wstd.close()

_runc = " ".join(['sofia_pipeline.py', sofia_file])

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types



# -*- coding: future_fstrings -*-
import stimela
import os
import sys
import unittest
import subprocess
import shutil
import glob
from stimela.exceptions import *
from stimela.dismissable import dismissable as sdm
from stimela.pathformatter import pathformatter as spf
import stimela.cargo as cargo

class TestBasicRecipe(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        DIR = os.path.dirname(__file__)
        global INPUT, MSDIR, OUTPUT, MS, PREFIX, LSM, MS_SIM
        INPUT = os.path.join(DIR, "input")
        MSDIR = "msdir"
        global OUTPUT
        OUTPUT = "/tmp/output"
        global CABPATH 
        CABPATH = os.path.join(DIR, "cab")

    @classmethod
    def tearDownClass(cls):
        pass

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        global MSDIR
        global OUTPUT
        if os.path.exists(MSDIR):
            shutil.rmtree(MSDIR)
        if os.path.exists(OUTPUT):
            shutil.rmtree(OUTPUT)
        if os.path.exists(INPUT):
            shutil.rmtree(INPUT)
        if os.path.exists("stimela_parameter_files"):
            shutil.rmtree("stimela_parameter_files")
        for log in glob.glob("log-*"):
            os.remove(log)

    def setUp(self):
        unittest.TestCase.setUp(self)
        if not os.path.isdir(INPUT):
            os.mkdir(INPUT)

    def test_define_cab(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("customcab", ms_dir=MSDIR)
        rrr.add("cab/custom", "test1", {
            "bla1": "a"
        }, 
        cabpath=CABPATH,
        input=INPUT, output=OUTPUT)
        assert len(rrr.jobs) == 1
        rrr.run() #validate and run
        assert rrr.jobs[0].job._cab.parameters[0].value == "a"
        assert len(rrr.completed) == 1
        assert len(rrr.remaining) == 0

    def test_invalid_choice(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("invchoice", ms_dir=MSDIR)
        rrr.add("cab/custom", "test1", {
        "bla1": "d" # only accepts a, b or c
        },
        cabpath=CABPATH,
        input=INPUT, output=OUTPUT)
        with self.assertRaises(PipelineException):
            rrr.run() #validate and run

    def test_dismissable(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("testdismissable", ms_dir=MSDIR)
        rrr.add("cab/custom", "test1", {
            "bla1": "a", # only accepts a, b or c
            "bla4": sdm("abc"),
            "bla3": sdm(None)
        }, 
        cabpath=CABPATH,
        input=INPUT, output=OUTPUT)
        rrr.run() #validate and run
        assert rrr.jobs[0].job._cab.parameters[0].value == "a"
        assert rrr.jobs[0].job._cab.parameters[1].value is None
        assert rrr.jobs[0].job._cab.parameters[2].value is None
        assert rrr.jobs[0].job._cab.parameters[3].value == ["abc"]

    def test_floattypefail(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("testfloattypefail", ms_dir=MSDIR)
        rrr.add("cab/custom", "test1", {
            "bla1": "a", # only accepts a, b or c
            "bla3": "1.0a",
        },
        cabpath=CABPATH,
        input=INPUT, output=OUTPUT)
        with self.assertRaises(PipelineException):
            rrr.run() #validate and run

    def test_floattypesuccess(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("testfloattypesuccess", ms_dir=MSDIR)
        rrr.add("cab/custom", "test1", {
            "bla1": "a", # only accepts a, b or c
            "bla3": 4.0,
        },
        cabpath=CABPATH,
        input=INPUT, output=OUTPUT)
        rrr.run() #validate and run
        assert rrr.jobs[0].job._cab.parameters[2].value == [4.0]

    def test_required(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("testrequired", ms_dir=MSDIR)
        rrr.add("cab/custom", "test1", {
            "bla3": 4.0,
        },
        cabpath=CABPATH,
        input=INPUT, output=OUTPUT)
        with self.assertRaises(PipelineException):
            rrr.run() #validate and run

    def test_iooverride(self):
        global MSDIR
        global INPUT
        global CABPATH
        with open(os.path.join(INPUT, "testinput.txt"), "w+") as f:
            pass
        global OUTPUT
        stimela.register_globals()
        rrr = stimela.Recipe("testiooverrides", ms_dir=MSDIR)
        rrr.add("cab/custom", "test1", {
            "bla1": "a", # only accepts a, b or c
            "bla2": "testinput.txt:input",
        },
        cabpath=CABPATH,
        input=INPUT, output=OUTPUT)
        rrr.run() #validate and run
        assert rrr.jobs[0].job._cab.parameters[0].value == "a"
        assert rrr.jobs[0].job._cab.parameters[1].value == os.path.join(rrr.jobs[0].job.IODEST["input"], 
                "testinput.txt")

    def test_iopathval(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("ioval", ms_dir=MSDIR)
        rrr.add("cab/custom", "test1", {
            "bla1": "a", # only accepts a, b or c
            "bla2": "testinput2.txt:input",
        },
        cabpath=CABPATH,
        input=INPUT, output=OUTPUT)
        with self.assertRaises(PipelineException): # not exist during validation
            rrr.run() #validate and run

    def test_iopathlist(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("pathlist", ms_dir=MSDIR)
        rrr.add("cab/custom", "test1", {
            "bla1": "a", # only accepts a, b or c
            "bla5": ["testinput2.txt:input",
                     "testinput3.txt:msfile",
                     spf("{}hello\{reim\}.fits,{}to.fits,{}world.fits", "input", "msfile", "output")],
        },
        cabpath=CABPATH,
        input=INPUT, output=OUTPUT)
        rrr.run() #validate and run
        assert rrr.jobs[0].job._cab.parameters[4].value[0] == os.path.join(rrr.jobs[0].job.IODEST["input"], 
                "testinput2.txt")
        assert rrr.jobs[0].job._cab.parameters[4].value[1] == os.path.join(rrr.jobs[0].job.IODEST["msfile"],
                "testinput3.txt")
        assert rrr.jobs[0].job._cab.parameters[4].value[2] == \
                "{}/hello{{reim}}.fits,{}/to.fits,{}/world.fits".format(rrr.jobs[0].job.IODEST["input"],
                    rrr.jobs[0].job.IODEST["msfile"],
                    rrr.jobs[0].job.IODEST["output"]
                    )


if __name__ == "__main__":
    unittest.main()



# -*- coding: future_fstrings -*-
import sys

from scabha import config, parse_parameters, prun

args = [config.binary] + parse_parameters(repeat=" ")

# run the command
if prun(args) != 0:
    sys.exit(1)




# -*- coding: future_fstrings -*-
import sys, os, os.path

from scabha import log, config, parameters, prun_multi, OUTPUT

ms = os.path.abspath(parameters.ms)
os.chdir(OUTPUT)

errors = prun_multi([f"{config.binary} {ms} {args}" for args in parameters.args])

for cmd, exc in errors:
    log.error(f"{cmd}: failed with return code {exc.returncode}")

if errors and not parameters.get('ignore_errors'):
    sys.exit(1)



import os
import sys
import yaml
import rfinder
import glob
import subprocess
import shlex
import shutil


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]
args = []
msname = None

pkg_path = os.path.dirname(os.path.realpath(rfinder.__file__))
rfinder_file = '{:s}/rfinder_default.yml'.format(pkg_path)

with open(rfinder_file) as f:
    list_doc = yaml.load(f)

list_doc['general']['outdir'] = '{:s}/'.format(OUTPUT)
list_doc['general']['workdir'] = '{:s}/'.format(MSDIR)

for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue

    if name == 'msname':
        list_doc['general']['msname'] = value.split('/')[-1]
        continue

    for key, val in list_doc.items():
        if type(val) == dict:
            for k1, v1 in val.items():
                if type(v1) == dict:
                    for k2, v2 in v1.items():
                        if k2 == name:
                            list_doc[key][k1][k2] = value
                else:
                    if k1 == name:
                        list_doc[key][k1] = value
        else:
            if key == name:
                list_doc[key] = value

edited_file = 'rfinder_default.yml'
with open(edited_file, "w") as f:
    yaml.dump(list_doc, f)


_runc = 'rfinder -c %s' % edited_file

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types




# -*- coding: future_fstrings -*-
import Crasa.Crasa as crasa
from scabha import config, parameters_dict, prun
from casacore.table import table
import os
import numpy

print(f"Running CASA task '{config.binary}'")

args = parameters_dict

task = crasa.CasaTask(config.binary, **args)
task.run()

gtab = args["caltable"]
if not os.path.exists(gtab):
    raise RuntimeError(f"The gaintable was not created. Please refer to CASA {config.binary} logfile for further details")

tab = table(gtab)
field_ids = numpy.unique(tab.getcol("FIELD_ID"))
tab.close()

tab = table(gtab+"::FIELD")
field_names = tab.getcol("NAME")
tab.close()

field_in = args["field"].split(",")

try:
    ids = list(map(int, field_in))
except ValueError:
    ids = list(map(lambda a: field_names.index(a), field_in))

if not set(ids).issubset(field_ids):
    raise RuntimeError(f"Some field(s) do not have solutions after the calibration. Please refer to CASA {config.binary} logfile for further details")



# -*- coding: future_fstrings -*-
import sys
from scabha import config, parameters_dict, prun, parse_parameters

"""
config: 
    contains the sections before parameters in params.json
    .binary has the name of the binary to be executed
parameters_dict: dict
    contains all the provided parameters, even the positional ones
parse_parameters: function
    Forms a list containing all the provided arguments for execution.
    This is a helper function that formats stuff so that you don't have to
prun: function
    Execute your binary with the provided arguments
"""

args = [config.binary] + parse_parameters(parameters_dict)
# run the command

if prun(args) != 0:
    sys.exit(1)





# -*- coding: future_fstrings -*-
import sys

from scabha import config, parse_parameters, prun

# If a list of fields is given, insert them as repeated arguments.
# Other arguments not allowed to be lists.
args = [config.binary] + parse_parameters(repeat=True,
                                          positional=["fitsq", "fitsu", "freqs"], mandatory=["fitsq", "fitsu", "freqs"])

# run the command
if prun(args) != 0:
    sys.exit(1)



# -*- coding: future_fstrings -*-
import stimela
import os
import sys
import unittest
import subprocess
import shutil
import glob
from stimela.exceptions import *
from stimela.dismissable import dismissable as sdm
from stimela.pathformatter import pathformatter as spf
from stimela import cargo, singularity

class TestBasicRecipe(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        global INPUT, MSDIR, OUTPUT, MS, PREFIX, LSM, MS_SIM
        INPUT = os.path.join(os.path.dirname(__file__), "input")
        MSDIR = "msdir"
        global OUTPUT
        OUTPUT = "/tmp/output"

        # Start stimela Recipe instance
        import stimela.main as main
        os.chdir(os.path.dirname(__file__))
        cab = cargo.cab.CabDefinition(parameter_file="cab/custom/parameters.json")
        global SINGULARITY, PODMAN
        SINGULARITY = False
        PODMAN = False
        if singularity.version:
            if singularity.BINARY_NAME == "singularity":
                SINGULARITY = singularity.version >= "2.6.0"
            else:
                SINGULARITY = True


    @classmethod
    def tearDownClass(cls):
        pass

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        global MSDIR
        global OUTPUT
        if os.path.exists(MSDIR):
            shutil.rmtree(MSDIR)
        if os.path.exists(OUTPUT):
            shutil.rmtree(OUTPUT)
        if os.path.exists(INPUT):
            shutil.rmtree(INPUT)
        if os.path.exists("stimela_parameter_files"):
            shutil.rmtree("stimela_parameter_files")
        for log in glob.glob("log-*"):
            os.remove(log)

    def setUp(self):
        unittest.TestCase.setUp(self)
        if not os.path.isdir(INPUT):
            os.mkdir(INPUT)

    def test_singularity(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global SINGULARITY
        
        if SINGULARITY is False:
            return

        stimela.register_globals()
        rrr = stimela.Recipe("singularitypaths",
                             ms_dir=MSDIR,
                             JOB_TYPE="singularity",
                             cabpath="cab/",
                             singularity_image_dir=os.environ["STIMELA_PULLFOLDER"],
                             log_dir="logs")
        rrr.add("cab/custom", "test1", {
            "bla1": "a", # only accepts a, b or c
            "bla5": ["testinput2.txt:input",
                     "testinput3.txt:msfile",
                     spf("{}hello\{reim\}.fits,{}to.fits,{}world.fits", "input", "msfile", "output")],
        }, input=INPUT, output=OUTPUT)
        rrr.run() #validate and run

        assert rrr.jobs[0].job._cab.parameters[4].value[0] == os.path.join(rrr.jobs[0].job.IODEST["input"], 
                    "testinput2.txt")
        assert rrr.jobs[0].job._cab.parameters[4].value[1] == os.path.join(rrr.jobs[0].job.IODEST["msfile"],
                    "testinput3.txt")
        assert rrr.jobs[0].job._cab.parameters[4].value[2] == \
                "{}/hello{{reim}}.fits,{}/to.fits,{}/world.fits".format(
                    rrr.jobs[0].job.IODEST["input"],
                    rrr.jobs[0].job.IODEST["msfile"],
                    rrr.jobs[0].job.IODEST["output"]
                )

    def test_podman(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global PODMAN
        if PODMAN is False:
            return
        stimela.register_globals()
        rrr = stimela.Recipe("podmanpaths",
                             ms_dir=MSDIR,
                             JOB_TYPE="podman",
                             cabpath="cab/",
                             log_dir="logs")
        rrr.add("cab/custom", "test1", {
            "bla1": "a", # only accepts a, b or c
            "bla5": ["testinput2.txt:input",
                     "testinput3.txt:msfile",
                     spf("{}hello\{reim\}.fits,{}to.fits,{}world.fits", "input", "msfile", "output")],
        }, input=INPUT, output=OUTPUT)
        rrr.run() #validate and run
        assert rrr.jobs[0].job._cab.parameters[4].value[0] == os.path.join(rrr.jobs[0].job.IODEST["input"], 
                    "testinput2.txt")
        assert rrr.jobs[0].job._cab.parameters[4].value[1] == os.path.join(rrr.jobs[0].job.IODEST["msfile"],
                    "testinput3.txt")
        assert rrr.jobs[0].job._cab.parameters[4].value[2] == \
                "{}/hello{{reim}}.fits,{}/to.fits,{}/world.fits".format(
                    rrr.jobs[0].job.IODEST["input"],
                    rrr.jobs[0].job.IODEST["msfile"],
                    rrr.jobs[0].job.IODEST["output"]
                )

if __name__ == "__main__":
    unittest.main()



# -*- coding: future_fstrings -*-
import stimela
import os
import unittest
import subprocess
import shutil


class TestKat7Reduce(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        # I/O
        global INPUT
        INPUT = 'input'
        global MSDIR
        MSDIR = 'msdir'

        global MS
        MS = 'kat-7-small.ms'
        global PREFIX
        PREFIX = 'kat7_small_LBand'

        # Fields
        global GCAL
        GCAL = 'PKS2326-477'
        global TARGET
        TARGET = '1'
        global BPCAL
        BPCAL = 'PKS1934-638'

        # Reference antenna
        global REFANT
        REFANT = '0'

        # Calibration tables
        global ANTPOS_TABLE
        ANTPOS_TABLE = PREFIX + '.antpos:output'
        global BPCAL_TABLE
        BPCAL_TABLE = PREFIX + '.B0:output'
        global DELAYCAL_TABLE
        DELAYCAL_TABLE = PREFIX + '.K0:output'
        global GAINCAL_TABLE
        GAINCAL_TABLE = PREFIX + '.G0:output'
        global FLUXSCALE_TABLE
        FLUXSCALE_TABLE = PREFIX + '.fluxscale:output'

        global LABEL
        LABEL = "test_reduction"
        global OUTPUT
        OUTPUT = "output_%s" % LABEL
        global MSCONTSUB
        MSCONTSUB = MS + '.contsub'

        global SPW
        SPW = '0:100~355'
        # Calibration tables
        global LSM0
        LSM0 = PREFIX + '.lsm.html'
        global SELFCAL_TABLE1
        SELFCAL_TABLE1 = PREFIX + '.SF1:output'
        global IMAGE1
        IMAGE1 = PREFIX + 'image1:output'
        global MASK1
        MASK1 = PREFIX + 'mask1.fits'
        global IMAGE2
        IMAGE2 = PREFIX + 'image2:output'
        global nchans
        nchans = 256
        global chans
        chans = [100, 355]

        # Clean-Mask-Clean
        global imname0
        imname0 = PREFIX + 'image0'
        global maskname0
        maskname0 = PREFIX + 'mask0.fits'
        global maskname01
        maskname01 = PREFIX + 'mask01.fits'
        global imname1
        imname1 = PREFIX + 'image1'
        global corr_ms
        corr_ms = MS + '-corr.ms'
        global lsm0
        lsm0 = PREFIX + '-LSM0'
        stimela.register_globals()

    @classmethod
    def tearDownClass(cls):
        unittest.TestCase.tearDownClass()

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    def setUp(self):
        unittest.TestCase.setUp(self)

    def test_end_to_end_reduction(self):
        global INPUT, OUTPUT, MSDIR, MS, LABEL
        global GAINCAL_TABLE2, FLUXSCALE_TABLE, GAINCAL_TABLE, DELAYCAL_TABLE, BPCAL_TABLE, ANTPOS_TABLE
        global REFANT, BPCAL, TARGET, GCAL, PREFIX
        global MSCONTSUB, SPW, LSM0, SELFCAL_TABLE1, corr_ms, lsm0
        global IMAGE1, IMAGE2, MASK1, nchans, chans, imname0, maskname0, maskname01, imname1

        recipe = stimela.Recipe('Test reduction script',
                                ms_dir=MSDIR, JOB_TYPE="docker", log_dir="logs")

        recipe.add('cab/casa_listobs', 'listobs', {
            "vis": MS
        },
            input=INPUT,
            output=OUTPUT,
            label='listobs:: some stats',
            time_out=300)

        recipe.add("cab/owlcat_plotelev", "plotobs", {
            "msname" : MS,
            "output-name" : "obsplot.png",
        },
            input=INPUT,
            output=OUTPUT,
            label="plotobs:: Plot elevation/azimuth vs LST/UTC")

        # It is common for the array to require a small amount of time to settle down at the start of a scan. Consequently, it has
        # become standard practice to flag the initial samples from the start
        # of each scan. This is known as 'quack' flagging
        recipe.add('cab/casa_flagdata', 'quack_flagging', {
            "vis": MS,
            "mode": 'quack',
            "quackinterval": 10.0,
            "quackmode": 'beg',
        },
            input=INPUT,
            output=OUTPUT,
            label='quack_flagging:: Quack flagging',
            time_out=300)

        # Flag the autocorrelations
        recipe.add("cab/politsiyakat_autocorr_amp", "flag_autopower", {
            "msname": MS,
            "field": "0,1,2",
            "cal_field": "0,2",
            "nrows_chunk": 5000,
            "scan_to_scan_threshold": 1.5,
            "antenna_to_group_threshold": 4,
            "nio_threads": 1,
            "nproc_threads": 32,
            },input=INPUT, output=OUTPUT, label="flag_autopower")

        recipe.add('cab/casa_flagdata', 'autocorr_flagging', {
            "vis":   MS,
            "mode":   'manual',
            "autocorr":   True,
        },
            input=INPUT,
            output=OUTPUT,
            label='autocorr_flagging:: Autocorrelations flagging',
            time_out=300)

        # Flag bad channels
        recipe.add('cab/casa_flagdata', 'badchan_flagging', {
            "vis":   MS,
            "mode":   'manual',
            "spw":   "0:113~113,0:313~313,0:369~369,0:601~607,0:204~204,0:212~212,0:594~600",

        },
            input=INPUT,
            output=OUTPUT,
            label='badchan_flagging:: Bad Channel flagging',
            time_out=300)

        recipe.add('cab/casa_clearcal', 'clearcal',
                   {
                       "vis":   MS,
                       "addmodel":   True
                   },
                   input=INPUT,
                   output=OUTPUT,
                   label='clearcal:: casa clearcal',
                   time_out=300)

        recipe.add('cab/casa_setjy', 'set_flux_scaling', {
            "vis":   MS,
            "field":   BPCAL,
            "standard":   'Perley-Butler 2010',
            "usescratch":   True,
            "scalebychan":   True,
        },
            input=INPUT,
            output=OUTPUT,
            label='set_flux_scaling:: Set flux density value for the amplitude calibrator',
            time_out=300)

        recipe.add('cab/casa_bandpass', 'bandpass_cal', {
            "vis": MS,
            "caltable": BPCAL_TABLE,
            "field": BPCAL,
            "refant": REFANT,
            "spw": SPW,
            "solint": 'inf',
            "bandtype": 'B',
            #                        "opacity"   : 0.0,
            #                        "gaincurve" : False,
        },
            input=INPUT,
            output=OUTPUT,
            label='bandpass_cal:: Bandpass calibration',
            time_out=300)

        recipe.add('cab/ragavi', 'ragavi_gains_plot_bandpass', {
                   'table': BPCAL_TABLE,
                   'gaintype': "B",
                   'htmlname': PREFIX + '_B0_amp_chan'
                   },
                   input=INPUT,
                   output=OUTPUT,
                   label='ragavi_gains_plot_bandpass:: Plot bandpass table',
                   time_out=1200
                   )

        # display the bandpass solutions. Note that in the plotcal inputs below, the amplitudes are being displayed as a function of
        # frequency channel. The parameter subplot=221 is used to display multiple plots per page (2 plots per page in the y
        # direction and 2 in the x direction). The first two commands below show the amplitude solutions (one per each polarization)
        # and the last two show the phase solutions (one per each polarization). Parameter iteration='antenna' is used to step
        # through separate plots for each antenna.
        recipe.add('cab/casa_plotcal', 'plot_bandpass_amp_R', {
            "caltable": BPCAL_TABLE,
            "poln": 'R',
            "xaxis": 'chan',
            "yaxis": 'amp',
            "field": BPCAL,
            "spw": SPW,
            "subplot": 221,
            "figfile": PREFIX + '-B0-R-amp.png',
        },
            input=INPUT,
            output=OUTPUT,
            label='plot_bandpass_amp_R:: Plot bandpass table. AMP, R',
            time_out=1200)

        # Gain calibration - amplitude and phase - first for BPCAL.
        recipe.add('cab/casa_gaincal', 'gaincal_bp', {
            "vis": MS,
            "caltable": GAINCAL_TABLE,
            "field": "{0:s},{1:s}".format(BPCAL, GCAL),
            "solint": 'inf',
            "refant": '',
            "gaintype": 'G',
            "calmode": 'ap',
            "spw": SPW,
            "solnorm": False,
            "gaintable": [BPCAL_TABLE],
            "interp": ['nearest'],
        },
            input=INPUT,
            output=OUTPUT,
            label="gaincal:: Gain calibration",
            time_out=300, 
            version=None)

        # Set fluxscale
        recipe.add('cab/casa_fluxscale', 'fluxscale', {
            "vis": MS,
            "caltable": GAINCAL_TABLE,
            "fluxtable": FLUXSCALE_TABLE,
            "reference": [BPCAL],
            "transfer": [GCAL],
            "save_result" : "fluxinfo.pickle",
            "incremental": False,
        },
            input=INPUT,
            output=OUTPUT,
            label='fluxscale:: Set fluxscale',
            time_out=300)

        # Apply calibration to BPCAL
        recipe.add('cab/casa_applycal', 'applycal_bp', {
            "vis": MS,
            "field": BPCAL,
            "gaintable": [BPCAL_TABLE, FLUXSCALE_TABLE],
            "gainfield": ['', '', BPCAL],
            "interp": ['', '', 'nearest'],
            "calwt": [False],
            "parang": False,
            "applymode": "calflag",
        },
            input=INPUT,
            output=OUTPUT,
            label='applycal_bp:: Apply calibration to Bandpass Calibrator',
            time_out=1800)

        # Flag the phase
        recipe.add("cab/politsiyakat_cal_phase", "flag_calphase", {
            "msname": MS,
            "field": ",".join(["0","1","2"]),
            "cal_field": ",".join(["0","2"]),
            "nrows_chunk": 5000,
            "data_column": "CORRECTED_DATA",
            "scan_to_scan_threshold": 1.5,
            "baseline_to_group_threshold": 4,
            "nio_threads": 1,
            "nproc_threads": 32,
            },input=INPUT, output=OUTPUT, label="flag_calphase")

        recipe.run()

        recipe = stimela.Recipe('KAT reduction script 2',
                                ms_dir=MSDIR, JOB_TYPE="docker", log_dir="logs")
        # Copy CORRECTED_DATA to DATA, so we can start uv_contsub
        recipe.add("cab/msutils", "move_corrdata_to_data", {
            "command": "copycol",
            "msname": MS,
            "fromcol": "CORRECTED_DATA",
            "tocol": "DATA",
        },
            input=INPUT, output=OUTPUT,
            label="move_corrdata_to_data::msutils",
            time_out=1800)

        os.system("rm -rf {}/{}-corr.ms".format(MSDIR, MS[:-3]))
        recipe.add('cab/casa_split', 'split_corr_data',
                   {
                       "vis":   MS,
                       "outputvis":   MS[:-3] + '-corr.ms',
                       "field":   str(BPCAL),
                       "spw":   SPW,
                       "datacolumn":   'data',
                   },
                   input=INPUT,
                   output=OUTPUT,
                   label='split_corr_data:: Split corrected data from MS',
                   time_out=1800)

        MS = MS[:-3] + '-corr.ms'

        recipe.add('cab/casa_clearcal', 'prep_split_data',
                   {
                       "vis":   MS,
                       "addmodel":   True
                   },
                   input=INPUT,
                   output=OUTPUT,
                   label='prep_split_data:: Prep split data with casa clearcal',
                   time_out=1800)

        # Clean-Mask-Clean
        imname0 = PREFIX + 'image0'
        maskname0 = PREFIX + 'mask0.fits'
        maskname01 = PREFIX + 'mask01.fits'
        imname1 = PREFIX + 'image1'

        recipe.add('cab/casa_tclean', 'image_target_field_r1', {
            "vis":   MS,
            "datacolumn":  "corrected",
            "field":   "0",
            "start":   21,  # Other channels don't have any data
            "nchan":   235 - 21,
            "width":   1,
            # Use Briggs weighting to weigh visibilities for imaging
            "weighting":   "briggs",
            "robust":   0,
            "imsize":   256,                   # Image size in pixels
            "cellsize":   "30arcsec",                      # Size of each square pixel
            "niter":   100,
            "stokes":   "I",
            "prefix":   '%s:output' % (imname1),
        },
            input=INPUT,
            output=OUTPUT,
            label="image_target_field_r1:: Image target field second round",
            time_out=300)

        recipe.add('cab/cleanmask', 'mask0', {
            "image": '%s.image.fits:output' % (imname1),
            "output": '%s:output' % (maskname0),
            "dilate": False,
            "sigma": 20,
        },
            input=INPUT,
            output=OUTPUT,
            label='mask0:: Make mask',
            time_out=1800)

        lsm0 = PREFIX + '-LSM0'
        # Source finding for initial model
        recipe.add("cab/pybdsm", "extract_init_model", {
            "image":  '%s.image.fits:output' % (imname1),
            "outfile":  '%s:output' % (lsm0),
            "thresh_pix":  25,
            "thresh_isl":  15,
            "port2tigger":  True,
        },
            input=INPUT, output=OUTPUT,
            label="extract_init_model:: Make initial model from preselfcal image",
            time_out=1800)

        # First selfcal round

        recipe.add("cab/calibrator", "calibrator_Gjones_subtract_lsm0", {
            "skymodel": "%s.lsm.html:output" % (lsm0),
            "msname": MS,
            "threads": 16,
            "column": "DATA",
            "output-data": "CORR_RES",
            "Gjones": True,
            # Ad-hoc right now, subject to change
            "Gjones-solution-intervals": [20, 0],
            "Gjones-matrix-type": "GainDiagPhase",
            "tile-size": 512,
            "field-id": 0,
        },
            input=INPUT, output=OUTPUT,
            label="calibrator_Gjones_subtract_lsm0:: Calibrate and subtract LSM0",
            time_out=1800)

        # Diversity is a good thing... lets add some DDFacet to this soup bowl
        imname = PREFIX + 'ddfacet'

        recipe.add("cab/ddfacet", "ddfacet_test",
                   {
                       "Data-MS": [MS],
                       "Output-Name": imname,
                       "Image-NPix": 256,
                       "Image-Cell": 30,
                       "Cache-Reset": True,
                       "Freq-NBand": 2,
                       "Weight-ColName": "WEIGHT",
                       "Data-ChunkHours": 10,
                       "Beam-FITSFeed": "rl",
                       "Data-Sort": True,
                       "Log-Boring": True,
                       "Deconv-MaxMajorIter": 1,
                       "Deconv-MaxMinorIter": 20,
                   },
                   input=INPUT, output=OUTPUT, shared_memory="8gb",
                   label="image_target_field_r0ddfacet:: Make a test image using ddfacet",
                   time_out=520)

        lsm1 = PREFIX + '-LSM0'
        # Source finding for initial model
        recipe.add("cab/pybdsm", "extract_init_model", {
            "image":  '%s.app.restored.fits:output' % (imname),
            "outfile":  '%s:output' % (lsm1),
            "thresh_pix":  25,
            "thresh_isl":  15,
            "port2tigger":  True,
        },
            input=INPUT, output=OUTPUT,
            label="extract_init_model:: Make initial model from preselfcal image",
            time_out=1800)

        # Stitch LSMs together
        lsm2 = PREFIX + '-LSM2'
        recipe.add("cab/tigger_convert", "stitch_lsms1", {
            "input-skymodel":   "%s.lsm.html:output" % lsm0,
            "output-skymodel":   "%s.lsm.html:output" % lsm2,
            "rename": True,
            "force": True,
            "append":   "%s.lsm.html:output" % lsm1,
        },
            input=INPUT, output=OUTPUT,
            label="stitch_lsms1::Create master lsm file",
            time_out=300)

        recipe.add("cab/calibrator", "calibrator_Gjones_subtract_lsm0", {
            "skymodel": "%s.lsm.html:output" % (lsm2),
            "msname": MS,
            "threads": 16,
            "column": "DATA",
            "output-data": "CORR_RES",
            "Gjones": True,
            # Ad-hoc right now, subject to change
            "Gjones-solution-intervals": [20, 0],
            "Gjones-matrix-type": "GainDiagPhase",
            "tile-size": 512,
            "field-id": 0,
        },
            input=INPUT, output=OUTPUT,
            label="calibrator_Gjones_subtract_lsm0:: Calibrate and subtract LSM0",
            time_out=1800)

        recipe.add('cab/casa_uvcontsub', 'uvcontsub',
                   {
                       "msname":    MS,
                       "field":    "0",
                       "fitorder":    1,
                   },
                   input=INPUT,
                   output=OUTPUT,
                   label='uvcontsub:: Subtract continuum in the UV plane',
                   time_out=1800)

        # Image HI
        recipe.add('cab/casa_clean', 'casa_dirty_cube',
                   {
                       "msname": MS + ".contsub",
                       "prefix": PREFIX,
                       "mode": 'channel',
                       "nchan": nchans,
                       "niter": 0,
                       "npix": 256,
                       "cellsize": 30,
                       "weight": 'natural',
                       "port2fits": True,
                   },
                   input=INPUT,
                   output=OUTPUT,
                   label='casa_dirty_cube:: Make a dirty cube with CASA CLEAN',
                   time_out=1800)

        recipe.add('cab/sofia', 'sofia',
                   {
                       #    USE THIS FOR THE WSCLEAN DIRTY CUBE
                       #    "import.inFile"     :   '{:s}-cube.dirty.fits:output'.format(combprefix),
                       #    USE THIS FOR THE CASA CLEAN CUBE
                       # CASA CLEAN cube
                       "import.inFile":   '{:s}.image.fits:output'.format(PREFIX),
                       "steps.doFlag": False,
                       "steps.doScaleNoise": True,
                       "steps.doSCfind": True,
                       "steps.doMerge": True,
                       "steps.doReliability": False,
                       "steps.doParameterise": False,
                       "steps.doWriteMask": True,
                       "steps.doMom0": False,
                       "steps.doMom1": False,
                       "steps.doWriteCat": True,
                       "flag.regions": [],
                       "scaleNoise.statistic": 'mad',
                       "SCfind.threshold": 4,
                       "SCfind.rmsMode": 'mad',
                       "merge.radiusX": 2,
                       "merge.radiusY": 2,
                       "merge.radiusZ": 2,
                       "merge.minSizeX": 2,
                       "merge.minSizeY": 2,
                       "port2tigger": False,
                       "merge.minSizeZ": 2,
                   },
                   input=INPUT,
                   output=OUTPUT,
                   label='sofia:: Make SoFiA mask and images',
                   time_out=1800)

        recipe.run()



# -*- coding: future_fstrings -*-
import os
import sys
import shlex
import configparser
import ast
from scabha import config, parameters_dict, prun

args = {}
parset = []

for name, value in parameters_dict.items():
    if value is None:
        continue
    elif value is False:
        value = 0
    elif value is True:
        value = 1
    elif name == 'parset':
        parset = [value]
        continue
    elif isinstance(value, list):
        value = ",".join(map(str, value))

    args[name] = value

# available jones terms
joneses = "g b dd".split()

try:
    soljones = args.pop("sol-jones")
except KeyError:
    conf = configparser.SafeConfigParser(inline_comment_prefixes="#")
    conf.read(parset[0])
    if 'jones' in conf.options('sol'):
        soljones = conf.get('sol', 'jones')
        if "[" in soljones:
            soljones = ast.literal_eval(soljones)
        else:
            soljones = soljones.split(",")
    else:
        soljones = ['g', 'de']
    if type(soljones) is list:
        soljones = ",".join(soljones)

for jones in joneses:
    if jones.lower() not in soljones.lower():
        jopts = filter(lambda a: a.startswith(
            "{0:s}-".format(jones)), args.keys())
        for item in list(jopts):
            del args[item]

opts = ["{0:s}sol-jones {1:s}".format(config.prefix, soljones)] + \
    ['{0}{1} {2}'.format(config.prefix, name, value)
     for name, value in args.items()]

_runc = " ".join([config.binary] + parset + opts)

argslist = shlex.split(_runc)

# run the command
if prun(argslist) != 0:
    sys.exit(1)



# -*- coding: future_fstrings -*-
import Crasa.Crasa as crasa
from scabha import config, parameters_dict, prun
from pyrap.tables import table
import os
import numpy

print(f"Running CASA task '{config.binary}'")

save_result = parameters_dict.pop("save_result", None)

task = crasa.CasaTask(config.binary, save_result=save_result, **parameters_dict)
task.run()

gtab = parameters_dict["caltable"]
if not os.path.exists(gtab):
    raise RuntimeError(f"The gaintable was not created. Please refer to CASA {config.binary} logfile for further details")

tab = table(gtab)
field_ids = numpy.unique(tab.getcol("FIELD_ID"))
tab.close()

field_in = parameters_dict["field"].split(",")
try:
    tab = table(gtab+"::FIELD")
    field_names = tab.getcol("NAME")
    tab.close()
except RuntimeError:
    # possible new table format
    # sadly Field name and Source name columns are empty
    # will need to figure this out, but ignoring the tests for now
    tab = table(gtab)
    field_names = numpy.unique(tab.getcol("FIELD_NAME"))
    tab.close()
    pass

if field_names:
    try:
        ids = list(map(int, field_in))
    except ValueError:
        ids = list(map(lambda a: field_names.index(a), field_in))
    if not set(ids).issubset(field_ids):
        raise RuntimeError(f"Some field(s) do not have solutions after the calibration. Please refer to CASA {config.binary} logfile for further details")




#config -*- coding: future_fstrings -*-
import sys

from scabha import config, parse_parameters, prun

args = [config.binary] + parse_parameters(repeat=" ")
for i in range(len(args)):
    if args[i] == '--verb':
        val = args.pop(i+1)
        if val == 'False':
            args.pop(i)
        
# run the command
if prun(args) != 0:
    sys.exit(1)




# -*- coding: future_fstrings -*-
import sys

from scabha import config, parse_parameters, prun

pars = parse_parameters(repeat=" ")
if "--psf-pars" in pars:
    pars = " ".join(pars).split()

args = [config.binary] + pars
    
# run the command
if prun(args) != 0:
    sys.exit(1)




#!/usr/bin/env python3
import os
import sys
import shlex
import shutil
import subprocess
import yaml
import glob


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]
args = []

for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue

    if name in ['recipe']:
        recipe = value
        continue

    if name in ['recipe-name']:
        recipe_name = value
        continue

    if name in ['support-files']:
        continue

    if name in ['step']:
        delimiter = ',' #param['delimiter']
        args += ['{0}{1} {2}'.format(cab['prefix'], name, delimiter.join(value))]

    if name in  ['params']:
        delimiter = ' ' #param['delimiter']
        args += ['{0}'.format(delimiter.join(value))]

_runc = " ".join([cab["binary"]] + [recipe, recipe_name] + args)

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



# -*- coding: future_fstrings -*-
import Crasa.Crasa as crasa
from scabha import config, parameters_dict, prun
from pyrap.tables import table
import os
import sys
import numpy
import astropy.io.fits as pyfits

args = parameters_dict

print(f"Running CASA task '{config.binary}'")

noise_image = args.pop('noise_image', False)
if noise_image:
    noise_sigma = args.pop('noise_sigma')
    noise_hdu = pyfits.open(noise_image)
    noise_data = noise_hdu[0].data
    noise_std = noise_data.std()
    threshold = noise_sigma*noise_std
    args['threshold'] = '{}Jy'.format(threshold)
else:
    args.pop('noise_sigma')

prefix = args['imagename']
port2fits = args.pop('port2fits', True)
keep_casa_images = args.pop("keep_casa_images", False)

task = crasa.CasaTask(config.binary, **args)
task.run()

nterms = args.get("nterms", 1)
images = ["flux", "model", "residual", "psf", "image"]
STD_IMAGES = images[:4]

convert = []
if port2fits:
    for image in images:
        img = "{:s}.{:s}".format(prefix, image)
        if image == 'flux':
            _images = [img]
        elif nterms > 1:
            _images = ["%s.tt%d" % (img, d) for d in range(nterms)]
            if image == "image":
                if nterms == 2:
                    alpha = img+".alpha"
                    alpha_err = img+".alpha.error"
                    _images += [alpha, alpha_err]
                if nterms == 3:
                    beta = img+".beta"
                    beta_err = img+".beta.error"
                    _images += [beta, beta_err]
        else:
            _images = [img]
        convert += _images

for _image in convert:
    sys.stdout.write(_image)
    if _image in STD_IMAGES and (not os.path.exists(_image)):
        raise RuntimeError(
            "Standard output from CASA clean task not found. Something went wrong durring cleaning, take a look at the logs and such")

    elif os.path.exists(_image):
        task = crasa.CasaTask(
            "exportfits", **dict(imagename=_image, fitsimage=_image+".fits", overwrite=True))
        task.run()

if not keep_casa_images:
    for _image in convert:
        os.system("rm -rf {}".format(_image))



import os
import sys
import Tigger
import numpy
import tempfile
import json
import codecs
import shlex
import shutil
import glob
import subprocess

from astLib.astWCS import WCS
from astropy.io.votable import parse_single_table
from Tigger.Models import SkyModel, ModelClasses


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

with codecs.open(CONFIG, "r", "utf8") as stdr:
    cab = json.load(stdr)

junk = cab["junk"]
args = []
msname = None

sofia_file = 'sofia_parameters.par'
wstd = open(sofia_file, 'w')

wstd.write('output.directory={:s}\n'.format(OUTPUT))
port2tigger = False
image = None

for param in cab['parameters']:
    name = param['name']
    value = param['value']
    dtype = param['dtype']

    # Fix the sofia issue of needing lowercase booleans.
    if dtype == 'bool':
        if (value == True) and (not name == 'port2tigger'):
            value = 'true'
        elif (not name == 'port2tigger'):
            value = 'false'

    if dtype == 'list:str' and value is not None:
        try:
            value = ",".join(map(str, value))  # Flatten the list into a string
        except ValueError:
            pass
    
    if value is None:
        continue
    if name == "port2tigger":
        port2tigger = value
        continue
    if name == "output.writeCatXML":
        writecat = value
    if name == "parameter.enable":
        parameterise = value
    if name == "input.data":
        image = value

    wstd.write('{0}={1}\n'.format(name, value))

wstd.close()

_runc = " ".join([cab['binary'], sofia_file])

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types

if not port2tigger:
    sys.exit(0)

# convert to data file to Tigger LSM
# First make dummy tigger model
tfile = tempfile.NamedTemporaryFile(suffix='.txt')
tfile.flush()

if image and writecat and parameterise:
    pass
else:
    sys.exit(0)

prefix = os.path.splitext(image)[0]
tname_lsm = prefix + ".lsm.html"
with open(tfile.name, "w") as stdw:
    stdw.write("#format:name ra_d dec_d i emaj_s emin_s pa_d\n")

model = Tigger.load(tfile.name)
tfile.close()


def tigger_src(src, idx):

    name = "SRC%d" % idx
    flux = ModelClasses.Polarization(src["f_sum"], 0, 0, 0)
    ra = numpy.deg2rad(src["ra"])
    dec = numpy.deg2rad(src["dec"])
    pos = ModelClasses.Position(ra, dec)
    ex = numpy.deg2rad(src["ell_maj"])
    ey = numpy.deg2rad(src["ell_min"])
    pa = numpy.deg2rad(src["ell_pa"])
    print(name)

    if ex and ey:
        shape = ModelClasses.Gaussian(ex, ey, pa)
    else:
        shape = None
    source = SkyModel.Source(name, pos, flux, shape=shape)
    # Adding source peak flux (error) as extra flux attributes for sources,
    # and to avoid null values for point sources I_peak = src["Total_flux"]
    if shape:
        source.setAttribute("I_peak", float(src["f_max"]))
    else:
        source.setAttribute("I_peak", float(src["f_sum"]))

    return source

table = parse_single_table('{0}_cat.xml'.format(prefix))
data = table.array

for i, src in enumerate(data):
    model.sources.append(tigger_src(src, i))

wcs = WCS(image)
centre = wcs.getCentreWCSCoords()
model.ra0, model.dec0 = map(numpy.deg2rad, centre)

model.save(tname_lsm)
# Rename using CORPAT
_runc = "tigger-convert %s --rename -f" % tname_lsm
try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)




# -*- coding: future_fstrings -*-
import Crasa.Crasa as crasa
from scabha import config, parameters_dict, prun
import os

print(f"Running CASA task '{config.binary}'")

save_result = parameters_dict.pop("save_result", None)
overwrite = parameters_dict.pop("overwrite", False)
fluxtable = parameters_dict['fluxtable']
if overwrite:
    os.system(f"rm -fr {fluxtable}")

task = crasa.CasaTask(config.binary, save_result=save_result, **parameters_dict)
task.run()



# -*- coding: future_fstrings -*-
import sys

from scabha import config, parameters, prun

args = [config.binary]

# convert arguments to flat list of PrefixName Arguments
for name, value in parameters.items():
    if value in [None, "", " ", False]:
        continue
    args.append(f'{config.prefix}{name}')

    if not isinstance(value, list):
        value = [value]

    args += list(map(str, value))

# run the command
if prun(args) != 0:
    sys.exit(1)




import os
import sys
import json
import yaml
import time
import tempfile
import inspect
import warnings
import re
import math
import codecs

class StimelaCabRuntimeError(RuntimeError):
    pass

class StimelaProcessRuntimeError(RuntimeError):
    pass



CPUS = 1

from .xrun_poll import xrun

def assign(key, value):
    frame = inspect.currentframe().f_back
    frame.f_globals[key] = value


def readJson(conf):
    with open(conf, "r") as _std:
        jdict = yaml.safe_load(_std)
        return jdict


def writeJson(config, dictionary):
    with codecs.open(config, 'w', 'utf8') as std:
        std.write(json.dumps(dictionary, ensure_ascii=False))


def get_Dockerfile_base_image(image):

    if os.path.isfile(image):
        dockerfile = image
    else:
        dockerfile = "{:s}/Dockerfile".format(image)

    with open(dockerfile, "r") as std:
        _from = ""
        for line in std.readlines():
            if line.startswith("FROM"):
                _from = line

    return _from


def change_Dockerfile_base_image(path, _from, label, destdir="."):
    if os.path.isfile(path):
        dockerfile = path
        dirname = os.path.dirname(path)
    else:
        dockerfile = "{:s}/Dockerfile".format(path)
        dirname = path

    with open(dockerfile, "r") as std:
        lines = std.readlines()
        for line in lines:
            if line.startswith("FROM"):
                lines.remove(line)

    temp_dir = tempfile.mkdtemp(
        prefix="tmp-stimela-{:s}-".format(label), dir=destdir)
    xrun(
        "cp", ["-r", "{:s}/Dockerfile {:s}/src".format(dirname, dirname), temp_dir])

    dockerfile = "{:s}/Dockerfile".format(temp_dir)

    with open(dockerfile, "w") as std:
        std.write("{:s}\n".format(_from))

        for line in lines:
            std.write(line)

    return temp_dir, dockerfile


def get_base_images(logfile, index=1):

    with open(logfile, "r") as std:
        string = std.read()

    separator = "[================================DONE==========================]"

    log = string.split(separator)[index-1]

    images = []

    for line in log.split("\n"):
        if line.find("<=BASE_IMAGE=>") > 0:
            tmp = line.split("<=BASE_IMAGE=>")[-1]
            image, base = tmp.split("=")
            images.append((image.strip(), base))

    return images


def substitute_globals(string, globs=None):
    sub = set(re.findall('\{(.*?)\}', string))
    globs = globs or inspect.currentframe().f_back.f_globals
    if sub:
        for item in map(str, sub):
            string = string.replace("${%s}" % item, globs[item])
        return string
    else:
        return False



# -*- coding: future_fstrings -*-
import Crasa.Crasa as crasa
from scabha import config, parameters_dict, prun

print(f"Running CASA task '{config.binary}'")

save_result = parameters_dict.pop("save_result", None)

task = crasa.CasaTask(config.binary, save_result=save_result, **parameters_dict)
task.run()



# -*- coding: future_fstrings -*-
import os
import sys
import argparse
from argparse import ArgumentParser
import textwrap as _textwrap
import signal
import stimela
from stimela import docker, singularity, podman, utils
from stimela.utils import logger
from stimela.cargo import cab

BASE = stimela.BASE
CAB = stimela.CAB
GLOBALS = stimela.GLOBALS
CAB_USERNAME = stimela.CAB_USERNAME

loglevels = "info debug error"

class MultilineFormatter(argparse.HelpFormatter):
    def _fill_text(self, text, width, indent):
        text = self._whitespace_matcher.sub(' ', text).strip()
        paragraphs = text.split('|n ')
        multiline_text = ''
        for paragraph in paragraphs:
            formatted_paragraph = _textwrap.fill(
                paragraph, width, initial_indent=indent, subsequent_indent=indent) + '\n\n'
            multiline_text = multiline_text + formatted_paragraph
        return multiline_text

def build(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv = ' ' + arg
    
    parser = ArgumentParser(description='Build executor (a.k.a cab) images')
    parser.add_argument("-b", "--base", action="store_true",
                        help="Build base images")

    parser.add_argument("-c", "--cab", metavar="CAB,CAB_DIR",
                        help="Executor image (name) name, location of executor image files")

    parser.add_argument("-uo", "--us-only",
                        help="Only build these cabs. Comma separated cab names")

    parser.add_argument("-i", "--ignore-cabs", default="",
                        help="Comma separated cabs (executor images) to ignore.")
    
    parser.add_argument("-p", "--podman", action="store_true",
        help="Build images using podman.")

    parser.add_argument("-nc", "--no-cache", action="store_true",
                        help="Do not use cache when building the image")


    jtype = "podman" if podman else "docker"
    args = parser.parse_args(argv)

    no_cache = ["--no-cache"] if args.no_cache else []

    if args.cab:
        raise SystemExit("DEPRECATION NOTICE: This feature has been deprecated. Please specify your \
                custom cab via the 'cabpath' option of the Recipe.add() function.")

    if args.base:
        # Build base and meqtrees images first
        BASE.remove("base")
        BASE.remove("meqtrees")
        BASE.remove("casa")
        BASE.remove("astropy")

        for image in ["base", "meqtrees", "casa", "astropy"] + BASE:
            dockerfile = "{:s}/{:s}".format(stimela.BASE_PATH, image)
            image = "stimela/{0}:{1}".format(image, stimela.__version__)
            #__call__(jtype).build(image,
            #             dockerfile, args=no_cache)

        return 0
    raise SystemExit("DEPRECATION NOTICE: The building of cab images has been deprecated")


def info(cabdir, header=False, display=True):
    """ prints out help information about a cab """

    # First check if cab exists
    pfile = "{}/parameters.json".format(cabdir)
    if not os.path.exists(pfile):
        raise RuntimeError("Cab could not be found at : {}".format(cabdir))
    # Get cab info
    cab_definition = cab.CabDefinition(parameter_file=pfile)
    if display:
        cab_definition.display(header)

    return cab_definition


def cabs(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv = ' ' + arg

    parser = ArgumentParser(description='List executor (a.k.a cab) images')
    parser.add_argument("-i", "--cab-doc",
                        help="Will display document about the specified cab. For example, \
to get help on the 'cleanmask cab' run 'stimela cabs --cab-doc cleanmask'")

    parser.add_argument("-l", "--list", action="store_true",
                        help="List cab names")

    parser.add_argument("-ls", "--list-summary", action="store_true",
                        help="List cabs with a summary of the cab")

    args = parser.parse_args(argv)

    if args.cab_doc:
        name = '{0:s}_cab/{1:s}'.format(CAB_USERNAME, args.cab_doc)
        cabdir = "{:s}/{:s}".format(stimela.CAB_PATH, args.cab_doc)
        info(cabdir)

    elif args.list_summary:
        for val in CAB:
            cabdir = "{:s}/{:s}".format(stimela.CAB_PATH, val)
            try:
                info(cabdir, header=True)
            except IOError:
                pass
    else:
        print(', '.join(CAB))


def run(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='Dockerized Radio Interferometric Scripting Framework.\n'
                                        'Sphesihle Makhathini <sphemakh@gmail.com>')

    add = parser.add_argument

    add("-r", "--repository", default="quay.io",
        help="Repository from which to pull docker images. The default repository is quay.io")

    add("-in", "--input",
        help="Input folder")

    add("-out", "--output",
        help="Output folder")

    add("-ms", "--msdir",
        help="MS folder. MSs should be placed here. Also, empty MSs will be placed here")

    add("-pf", "--pull-folder",
        help="Folder to store singularity images.")

    add("script",
        help="Run script")

    add("-g", "--globals", metavar="KEY=VALUE[:TYPE]", action="append", default=[],
        help="Global variables to pass to script. The type is assumed to string unless specified")

    add("-jt", "--job-type", choices=["docker", "singularity", "podman"],
        help="Container technology to use when running jobs")

    add("-ll", "--log-level", default="INFO", choices=loglevels.upper().split() + loglevels.split(),
        help="Log level. set to DEBUG/debug for verbose logging")

    args = parser.parse_args(argv)

    _globals = dict(_STIMELA_INPUT=args.input, _STIMELA_OUTPUT=args.output,
                    _STIMELA_MSDIR=args.msdir,
                    _STIMELA_JOB_TYPE=args.job_type,
                    _STIMELA_REP=args.repository,
                    _STIMELA_LOG_LEVEL=args.log_level.upper(),
                    _STIMELA_PULLFOLDER=args.pull_folder)

    args.job_type = args.job_type or "docker"
    nargs = len(args.globals)

    global GLOBALS

    if nargs:
        for arg in args.globals:
            if arg.find("=") > 1:
                key, value = arg.split("=")

                try:
                    value, _type = value.split(":")
                except ValueError:
                    _type = "str"

                GLOBALS[key] = eval("{:s}('{:s}')".format(_type, value))

    utils.CPUS = 1

    with open(args.script, 'r') as stdr:
        exec(stdr.read(), _globals)


def pull(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='Pull docker stimela base images')

    add = parser.add_argument

    add("-r", "--repository", default="quay.io",
        help="Repository from which to pull docker images. The default repository is quay.io")
        
    add("-im", "--image", nargs="+", metavar="IMAGE[:TAG]",
        help="Pull base image along with its tag (or version). Can be called multiple times")

    add("-f", "--force", action="store_true",
        help="force pull if image already exists")

    add("-s", "--singularity", action="store_true",
        help="Pull base images using singularity."
        "Images will be pulled into the directory specified by the enviroment varaible, STIMELA_PULLFOLDER. $PWD by default")

    add("-d", "--docker", action="store_true",
        help="Pull base images using docker.")

    add("-p", "--podman", action="store_true",
        help="Pull base images using podman.")

    add("-cb", "--cab-base", nargs="+",
        help="Pull base image for specified cab")

    add("-at", "--all-tags", action="store_true",
        help="Pull all tags for this image")

    add("-pf", "--pull-folder",
        help="Images will be placed in this folder. Else, if the environmnental variable 'STIMELA_PULLFOLDER' is set, then images will be placed there. "
        "Else, images will be placed in the current directory")

    args = parser.parse_args(argv)

    if args.pull_folder:
        pull_folder = args.pull_folder
    else:
        try:
            pull_folder = os.environ["STIMELA_PULLFOLDER"]
        except KeyError:
            pull_folder = "."

    if args.podman:
        jtype = "podman"
    elif args.singularity:
        jtype = "singularity"
    elif args.docker:
        jtype = "docker"
    else:
        jtype = "docker"

    images_ = []
    repository_ = []
    for cab in args.cab_base or []:
        if cab in CAB:
            filename = "/".join([stimela.CAB_PATH, cab, "parameters.json"])
            param = utils.readJson(filename)
            tags = param["tag"]
            if not isinstance(tags, list):
                tags = [tags]
            for tag in tags:
                images_.append(":".join([param["base"], tag]))
                repository_.append(param["hub"] if "hub" in param.keys() else args.repository)

    repository_ = repository_ + ([args.repository] * len(args.image) if args.image is not None else [])
    args.image = images_ + (args.image if args.image is not None else [])
    assert len(args.image) == len(repository_)
    if args.image:
        for hub, image in zip(repository_, args.image):
            if hub == "docker" or hub == "docker.io":
                hub = ""
            if args.singularity:
                simage = image.replace("/", "_")
                simage = simage.replace(":", "_") + singularity.suffix
                image = "/".join([hub, image]) if hub != "" else image
                singularity.pull(
                    image, simage, directory=pull_folder, force=args.force)
            else:
                if args.podman:
                    podman.pull("/".join([hub, image]) if hub != "" else image, force=args.force)
                else:
                    docker.pull("/".join([hub, image]) if hub != "" else image, force=args.force)
    else: # no cab bases or images specifically being pulled
        base = []
        repository_ = []
        for cab_ in CAB:
            filename = "/".join([stimela.CAB_PATH, cab_, "parameters.json"])
            param = utils.readJson(filename)

            cabdir = "{:s}/{:s}".format(stimela.CAB_PATH, cab_)
            _cab = info(cabdir, display=False)
            tags = _cab.tag
            if isinstance(tags, list):
                if not args.all_tags:
                    tags = [tags[-1]]
            else:
                tags = [tags]

            for tag in tags:
                base.append(f"{_cab.base}:{tag}")
                repository_.append(param["hub"] if "hub" in param.keys() else args.repository)

        assert len(base) == len(repository_)
        uniq_pulls = []
        for hub, image in zip(repository_, base):
            if hub == "docker" or hub == "docker.io":
                hub = ""
            if args.singularity:
                simage = image.replace("/", "_")
                simage = simage.replace(":", "_") + singularity.suffix
                image = "/".join([hub, image]) if hub != "" else image
                if image not in uniq_pulls:
                    uniq_pulls.append(image)
                    singularity.pull(
                        image, simage, directory=pull_folder, force=args.force)
            else:
                image = "/".join([hub, image]) if hub != "" else image
                if image not in uniq_pulls:
                    uniq_pulls.append(image)
                    if args.podman:
                        podman.pull(image, force=args.force)
                    else:
                        docker.pull(image, force=args.force)


def main(argv):
    for i, arg in enumerate(argv):
        if (arg[0] == '-') and arg[1].isdigit():
            argv[i] = ' ' + arg

    parser = ArgumentParser(description='Stimela: Dockerized Radio Interferometric Scripting Framework. '
                            '|n version {:s} |n install path {:s} |n '
                            'Sphesihle Makhathini <sphemakh@gmail.com>'.format(stimela.__version__,
                                                                               os.path.dirname(__file__)),
                            formatter_class=MultilineFormatter,
                            add_help=False)

    add = parser.add_argument

    add("-h", "--help",  action="store_true",
        help="Print help message and exit")

    add("-v", "--version", action='version',
        version='{:s} version {:s}'.format(parser.prog, stimela.__version__))

    add("command", nargs="*", metavar="command [options]",
        help="Stimela command to execute. For example, 'stimela help run'")

    options = []
    commands = dict(pull=pull, build=build, run=run, cabs=cabs)
#                    images=images, ps=ps,
#                    containers=containers, kill=kill,
#                    clean=clean)

    command = "failure"

    for cmd in commands:
        if cmd in argv:
            command = cmd

            index = argv.index(cmd)
            options = argv[index+1:]
            argv = argv[:index+1]

    args = parser.parse_args(argv)

    # Command is help and no other commands following
    try:
        main_help = (args.command[0] == "help" and len(args.command) == 1)
    except IndexError:
        main_help = True

    if args.help or main_help:
        parser.print_help()

        print("""
Run a command. These can be:

help    : Prints out a help message about other commands
build   : Build a set of stimela images
pull    : pull a stimela base images
run     : Run a stimela script
cabs    : Manage cab images

""")

        raise SystemExit

    # Separate commands into command and arguments
    cmd, argv = args.command[0], args.command[1:]

    # If we've got past the if statement above, and help
    # is the command then assume that help on a command
    # is requested
    if cmd == "help":
        # Request help on the sub-command
        cmd, argv = argv[0], ["-h"]
    else:
        argv = options

    # Get the function to execute for the command
    try:
        _cmd = commands[cmd]
    except KeyError:
        raise KeyError("Command '{:s}' not recognized "
                       "Run : 'stimela help' for help".format(cmd))

    # Invoke the command
    _cmd(argv)


def cli():
    main(sys.argv[1:])



import os
import sys
import subprocess
import yaml
import glob
import shutil
import shlex

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]

mslist = []
field = []

args = []
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value in [False, None]:
        continue
    if name == "primary-beam":
        value = "'{}'".format(value)
    if name == 'pa-range' and hasattr(value, '__iter__'):
        value = ','.join(value)
    if value is True:
        value = ""
    if name == 'pa-from-ms' and hasattr(value, '__iter__'):
        mslist = value
        continue
    if name == 'field-id'and hasattr(value, '__iter__'):
        field = value
        continue

    # Positional arguments
    if name == 'input-skymodel':
        inlsm = value
        continue
    elif name == 'output-skymodel':
        outlsm = value
        continue

    args.append('{0}{1} {2}'.format(cab['prefix'], name, value))

if mslist:
    if len(field) == 0:
        field = [0]*len(mslist)
    pa_from_ms = ','.join(['{0}:{1}'.format(ms, i)
                           for ms, i in zip(mslist, field)])
    args.append('--pa-from-ms {}'.format(pa_from_ms))

_runc = " ".join([cab['binary']] + args + [inlsm, outlsm])

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import numpy
import os
import sys
from pyrap.tables import table
import yaml
import math
import Cattery
from scabha import config, parameters_dict, prun

CODE = os.path.join(os.environ["STIMELA_MOUNT"], "code")
CONFIG = os.environ["CONFIG"]

def compute_vis_noise(msname, sefd, spw_id=0):
    """Computes nominal per-visibility noise"""
    tab = table(msname)
    spwtab = table(msname + "/SPECTRAL_WINDOW")

    freq0 = spwtab.getcol("CHAN_FREQ")[spw_id, 0]
    wavelength = 300e+6/freq0
    bw = spwtab.getcol("CHAN_WIDTH")[spw_id, 0]
    dt = tab.getcol("EXPOSURE", 0, 1)[0]
    dtf = (tab.getcol("TIME", tab.nrows()-1, 1)-tab.getcol("TIME", 0, 1))[0]

    # close tables properly, else the calls below will hang waiting for a lock...
    tab.close()
    spwtab.close()

    print(">>> %s freq %.2f MHz (lambda=%.2fm), bandwidth %.2g kHz, %.2fs integrations, %.2fh synthesis" % (
        msname, freq0*1e-6, wavelength, bw*1e-3, dt, dtf/3600))
    noise = sefd/math.sqrt(abs(2*bw*dt))
    print(">>> SEFD of %.2f Jy gives per-visibility noise of %.2f mJy" %
          (sefd, noise*1000))

    return noise

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

_params = parameters_dict

params = {}
options = {}
for name, value in _params.items():

    if value is None:
        continue

    params[name] = value

options["ms_sel.ddid_index"] = params.get('spw-id', 0)
options["ms_sel.field_index"] = params.get('field-id', 0)

tdlconf = params.get("tdlconf", None) or CODE+"/tdlconf.profiles"
section = params.get("section", None) or 'sim'
mode = params.pop("mode", None) or "simulate"
threads = params.pop("threads", 4)
skymodel = params.pop("skymodel", None)
beam_files_pattern = params.pop("beam-files-pattern", False)

if isinstance(skymodel, str):
    if os.path.exists(skymodel):
        options['me.sky.tiggerskymodel'] = 1
        options["tiggerlsm.filename"] = skymodel
    else:
        raise RuntimeError("ABORT: Could not find the skymodel")

modes = {
    "simulate":   'sim only',
    "add":   "add to MS",
    "subtract":   'subtract from MS',
}

column = params.pop("column", "CORRECTED_DATA")
incol = params.pop("input-column", "CORRECTED_DATA")

msname = params['msname']
options["ms_sel.msname"] = msname
options["sim_mode"] = modes[mode]
options["ms_sel.input_column"] = incol
options["ms_sel.output_column"] = column
options["me.use_smearing"] = 1 if params.pop('smearing', False) else 0
saveconf = params.pop('save-config', None)

addnoise = params.pop("addnoise", False)
options["ms_sel.tile_size"] = params.pop("tile-size", 16)

if addnoise:
    noise = params.pop("noise", 0) or compute_vis_noise(
        msname, params.pop("sefd", 551))
    options["noise_stddev"] = noise

gjones = params.pop("Gjones", False)
if gjones:
    gain_opts = {
        "me.g_enable":   1,
        "oms_gain_models.err-gain.error_model":   "SineError",
        "oms_gain_models.err-gain.max_period": params.pop("gain-max-period", 2),
        "oms_gain_models.err-gain.maxval": params.pop("gain-max-error", 1.2),
        "oms_gain_models.err-gain.min_period": params.pop("gain-min-period", 1),
        "oms_gain_models.err-gain.minval": params.pop("gain-min-error", 0.8),
    }

    options.update(gain_opts)

    phase_opts = {
        "oms_gain_models.err-phase.error_model": "SineError",
        "oms_gain_models.err-phase.max_period": params.pop("phase-max-period", 2),
        "oms_gain_models.err-phase.maxerr": params.pop("phase-max-error", 30.0),
        "oms_gain_models.err-phase.min_period": params.pop("phase-min-period", 1),
        "oms_gain_models.err-phase.minval": params.pop("phase-min-error", 5),
    }

    options.update(phase_opts)

beam = params.pop("Ejones", False)
if beam and beam_files_pattern:
    beam_opts = {
        "me.e_enable": 1,
        "me.p_enable": 1,
        "me.e_module": "Siamese_OMS_pybeams_fits",
        "pybeams_fits.sky_rotation":  1 if params.pop('parallactic-angle-rotation', False) else 0,
        "me.e_all_stations": 1,
        "pybeams_fits.l_axis": params.pop("beam-l-axis", "L"),
        "pybeams_fits.m_axis": params.pop("beam-m-axis", "M"),
        "pybeams_fits.filename_pattern": "'{}'".format(beam_files_pattern),
    }
    options.update(beam_opts)

    rms_perr = params.get("pointing-accuracy", 0)
    # Include pointing errors if needed
    if rms_perr:
        anttab = table(msname + "/" + "ANTENNA")
        NANT = anttab.nrows()

        options["me.epe_enable"] = 1
        perr = numpy.random.randn(
            NANT)*rms_perr, numpy.random.randn(NANT)*rms_perr
        ll, mm = " ".join(map(str, perr[0])), " ".join(map(str, perr[-1]))
        options['oms_pointing_errors.pe_l.values_str'] = "'%s'" % ll
        options['oms_pointing_errors.pe_m.values_str'] = "'%s'" % mm

field_center = params.pop("field-center", None)
if field_center and skymodel:
    if field_center.lower() == "ms":
        ftab = table(msname+"/FIELD")
        ra, dec = ftab.getcol("PHASE_DIR")[params.get('field-id', 0)][0]
        field_center = "J2000,%frad,%frad" % (ra, dec)
    tmp = "recentered_"+os.path.basename(skymodel)

    prun(["tigger-convert", "--recenter", field_center, skymodel, tmp, "-f"])
    options["tiggerlsm.filename"] = tmp

prefix = ['-s {}'.format(saveconf) if saveconf else ''] + \
    ["--mt {0} -c {1} [{2}]".format(threads, tdlconf, section)]
CATTERY_PATH = os.path.dirname(Cattery.__file__)
suffix = ["%s/Siamese/turbo-sim.py =_simulate_MS" % CATTERY_PATH]

args = []
for key, value in options.items():
    if isinstance(value, str) and value.find(' ') > 0:
        value = '"{:s}"'.format(value)
    args.append('{0}={1}'.format(key, value))

_runc = " ".join([config.binary] + prefix + args + suffix)

exitcode = prun(_runc)
if exitcode != 0:
    raise RuntimeError(f"Meqtrees failed with exit code {exitcode}. See logs for more details")



import json
import glob
from collections import OrderedDict

for cabfile in glob.glob("casa_*/*.json"):
    print(cabfile)
    with open(cabfile, 'rb') as stdr:
        cabdict = json.load(stdr, object_pairs_hook=OrderedDict)
    cabdict["version"] = ["4.7.2", "5.6.1-8", "5.8.0"]
    cabdict["tag"] = ["0.3.0-2", "1.6.3", "1.7.1"]
    cabdict["junk"] = ["%s.last" % (cabdict["binary"])]
    with open(cabfile, 'w') as stdw:
        json.dump(cabdict, stdw, indent=2)




# -*- coding: future_fstrings -*-
import sys

from scabha import config, parse_parameters, prun

# If a list of fields is given, insert them as repeated arguments.
# Other arguments not allowed to be lists.
args = [config.binary] + parse_parameters(repeat=True)

# run the command
if prun(args) != 0:
    raise SystemError(f" {config.binary} exited with a non-zero code. See logs for details")



# -*- coding: future_fstrings -*-
import stimela
from stimela.pathformatter import pathformatter as spf
import os
import unittest
import subprocess
import shutil


class TestMkReduce(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        # I/O
        global INPUT
        INPUT = 'input'
        global MSDIR
        MSDIR = 'msdir'

        global MS
        MS = '1491291289.1GC.ms'
        global PREFIX
        PREFIX = 'deep2'

        global LABEL
        LABEL = "test_mkreduction"
        global OUTPUT
        OUTPUT = "output_%s" % LABEL

        stimela.register_globals()

    @classmethod
    def tearDownClass(cls):
        unittest.TestCase.tearDownClass()
        #global OUTPUT
        # shutil.rmtree(OUTPUT)

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    def setUp(self):
        unittest.TestCase.setUp(self)

    def test_end_to_end_reduction(self):
        global INPUT, OUTPUT, MSDIR, MS, LABEL
        recipe = stimela.Recipe('Test reduction script',
                                ms_dir=MSDIR, JOB_TYPE="docker", log_dir="logs")

        imname1 = "deep2.1gc"
        imname2 = "deep2.2gc"
        recipe.add("cab/ddfacet", "ddfacet_test1",
                   {
                       "Data-MS": [MS],
                       "Output-Name": imname1,
                       "Image-NPix": 2048,
                       "Image-Cell": 2,
                       "Cache-Reset": True,
                       "Freq-NBand": 2,
                       "Freq-NDegridBand": 4,
                       "Weight-ColName": "WEIGHT",
                       "Data-ChunkHours": 0.1,
                       "Data-Sort": True,
                       "Log-Boring": True,
                       "Deconv-MaxMajorIter": 2,
                       "Deconv-MaxMinorIter": 1500,
                       "Predict-ColName": "MODEL_DATA"
                   },
                   input=INPUT, output=OUTPUT, shared_memory="8gb",
                   label="image1",
                   time_out=1800)

        recipe.add('cab/tricolour', "flag_data",
                   {
                       "ms": MS,
                       "data-column": "DATA",
                       "window-backend": 'numpy',
                       "flagging-strategy": "total_power",
                       "subtract-model-column": "MODEL_DATA",
                   },
                   input=INPUT, output=OUTPUT, label="flag_data",
                   time_out=1800)

        maskname0 = "MASK.fits"
        recipe.add('cab/cleanmask', 'mask0', {
            "image": '%s.app.restored.fits:output' % (imname1),
            "output": '%s:output' % (maskname0),
            "dilate": False,
            "sigma": 25,
        },
            input=INPUT,
            output=OUTPUT,
            label='mask0:: Make mask',
            time_out=1800)

        recipe.add("cab/ddfacet", "ddfacet_test2",
                   {
                       "Data-MS": [MS],
                       "Output-Name": imname1,
                       "Output-Images": "DdPAMRrIikze",
                       "Image-NPix": 4096,
                       "Image-Cell": 2.0,
                       "Cache-Reset": True,
                       "Freq-NBand": 2,
                       "Freq-NDegridBand": 4,
                       "Mask-External": '%s:output' % (maskname0),
                       "Weight-ColName": "WEIGHT",
                       "Data-ChunkHours": 0.1,
                       "Data-Sort": True,
                       "Log-Boring": True,
                       "Deconv-MaxMajorIter": 2,
                       "Deconv-MaxMinorIter": 1500,
                   },
                   input=INPUT, output=OUTPUT, shared_memory="24gb",
                   label="image2",
                   time_out=1800)

        recipe.add("cab/shadems", "shadems_test",
                   {
                       'ms': MS,
                       'xaxis': 'DATA:imag',
                       'yaxis': 'real',
                       'col': 'DATA',
                       'png': '%s_shadems_test_real_imag' % (PREFIX)
                   },
                   input=INPUT, output=OUTPUT,
                   label="shadems_test",
                   time_out=1800)

        # # First selfcal round
        recipe.add("cab/catdagger", "auto_tagger_{}_{}".format("decaltest", "1"), {
            'ds9-reg-file': "{}.{}.dE.reg".format("decaltest", "1"),
            'ds9-tag-reg-file': "{}.{}.dE.clusterleads.reg".format("decaltest", "1"),
            'sigma': 10,
            'min-distance-from-tracking-centre': 350,
            'noise-map': "{}.app.residual.fits:output".format(imname1),
        }, input=INPUT, output=OUTPUT, label="auto_tagger_{}_{}".format("decaltest", "1"), shared_memory="250g")

        recipe.add("cab/cubical_ddf", "cubical_cal",
                   {
                       'data-ms': MS,
                       'data-column': "DATA",
                       'dist-nworker': 4,
                       'dist-nthread': 1,
                       'dist-max-chunks': 20,
                       'data-freq-chunk': 0,
                       'data-time-chunk': 1,
                       'model-list': spf("MODEL_DATA+-{{}}{}@{{}}{}:{{}}{}@{{}}{}".format(
                           imname1+".DicoModel", "{}.{}.dE.reg".format("decaltest", "1"),
                           imname1+".DicoModel", "{}.{}.dE.reg".format("decaltest", "1")),
                           "output", "output", "output", "output"),
                       'log-verbose': "solver=0",
                       'weight-column': "WEIGHT",
                       'flags-apply': "FLAG",
                       'flags-auto-init': "legacy",
                       'madmax-enable': False,
                       'madmax-threshold': [0, 0, 10],
                       'madmax-global-threshold': [0, 0],
                       'sol-jones': 'g,dd',
                       'sol-stall-quorum': 0.95,
                       'out-name': "cubicaltest",
                       'out-column': "CORRECTED_DATA",
                       'log-verbose': "solver=0",
                       'g-type': "complex-2x2",
                       'g-freq-int': 0,
                       'dd-freq-int': 0,
                       'g-time-int': 20,
                       'dd-time-int': 20,
                       'g-max-iter': 10,
                       'sol-term-iters': 10,
                       'g-update-type': "phase-diag",
                       'dd-update-type': "complex-2x2",
                       'out-subtract-dirs': '1:',
                       'dd-fix-dirs': "0",
                       'dd-max-iter': 200,
                       'dd-clip-high': 0,
                       'dd-clip-low': 0,
                       'dd-max-prior-error': 0.35,
                       'dd-max-post-error': 0.35,
                       'degridding-NDegridBand': 3,
                       'degridding-MaxFacetSize': 0.15,
                       'out-mode': "sr",
                   }, input=INPUT, output=OUTPUT,
                   label="cubical",
                   shared_memory="24gb",
                   time_out=3600)

        recipe.add("cab/cubical", "cubical_cal2",
                   {
                       'data-ms': MS,
                       'data-column': "DATA",
                       'dist-nworker': 4,
                       'dist-nthread': 1,
                       'dist-max-chunks': 20,
                       'data-freq-chunk': 0,
                       'data-time-chunk': 1,
                       'model-list': spf("MODEL_DATA"),
                       'weight-column': "WEIGHT",
                       'flags-apply': "FLAG",
                       'flags-auto-init': "legacy",
                       'madmax-enable': False,
                       'madmax-threshold': [0, 0, 10],
                       'madmax-global-threshold': [0, 0],
                       'sol-jones': 'g',
                       'sol-stall-quorum': 0.95,
                       'out-name': "cubicaltest",
                       'out-column': "CORRECTED_DATA",
                       'log-verbose': "solver=2",
                       'g-type': "complex-2x2",
                       'g-freq-int': 0,
                       'g-time-int': 20,
                       'g-max-iter': 10,
                       'sol-term-iters': 10,
                       'out-overwrite' : True,
                       'g-update-type': "complex-2x2",

                   }, input=INPUT, output=OUTPUT,
                   label="cubical",
                   shared_memory="24gb",
                   time_out=1800)


        recipe.add("cab/ragavi_vis", "ragavi_vis_test",
                   {
                       'ms': MS,
                       'xaxis': 'imaginary',
                       'yaxis': 'real',
                       'data-column': 'CORRECTED_DATA',
                       'htmlname': "%s_ragavi_vis_real_imag" % (PREFIX)
                   },
                   input=INPUT, output=OUTPUT,
                   label="ragavi_vis_test",
                   time_out=1800)
        recipe.run()
